--
-- Wearable
--
-- @author Manuel Leithner
-- @date 30/07/14
--
-- Copyright (C) GIANTS Software GmbH, Confidential, All Rights Reserved.
--

---This is the specialization for all washable vehicles
-- @category Specializations
Wearable = {};

function Wearable.prerequisitesPresent(specializations)
	return true;
end;

---Called before loading
-- @param table savegame savegame
-- @includeCode
function Wearable:preLoad(savegame)
	self.setDirtAmount = Wearable.setDirtAmount;
	self.getDirtAmount = Wearable.getDirtAmount;
	self.getDirtMultiplier = Wearable.getDirtMultiplier;
	self.addWashingTrigger = Wearable.addWashingTrigger;
	self.removeWashingTrigger = Wearable.removeWashingTrigger;
	self.addWashableNode = Wearable.addWashableNode;
	self.removeWearableNode = Wearable.removeWearableNode;
end;

---Called on loading
-- @param table savegame savegame
-- @includeCode
function Wearable:load(savegame)
	self.sentDirtAmount = 0;
	self.washableDirtyFlag = self:getNextDirtyFlag();
	self.dirtAmount = 0;
	
	if hasXMLProperty(self.xmlFile, "vehicle.washable#dirtDuration") then
		self.washableNodes = {};
		self.dirtDuration = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.washable#dirtDuration"), 1) * 60 * 1000;
		if self.dirtDuration ~= 0 then
			self.dirtDuration = 1 / self.dirtDuration; -- dirtDuration == 0 disables dirt effect
		end;
		self.washDuration = math.max(Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.washable#washDuration"), 1) * 60 * 1000, 0.00001); -- washDuration == 0 washes vehicle immediately
		self.workMultiplier = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.washable#workMultiplier"), 1);
		self.fieldMultiplier = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.washable#fieldMultiplier"), 2);
	end;
	
	self.washingTriggers = {};
	
	self.scratchAmount = 0;
	self.scratchMultiplier = 1;
	self.scratchDuration = 1 / (Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.storeData.lifetime") / 10, 300) * 24 * 3600000);
end;

---Called after loading
-- @param table savegame savegame
-- @includeCode
function Wearable:postLoad(savegame)
	-- getting als washable nodes in postLoad to make sure also linked nodes are washable
	if self.washableNodes ~= nil then
		for _, comp in pairs(self.components) do
			self:addWashableNode(comp.node);
		end;
		
		if savegame ~= nil and Wearable.getIntervalMultiplier() ~= 0 then
			self:setDirtAmount(Utils.getNoNil(getXMLFloat(savegame.xmlFile, savegame.key.."#dirtAmount"), 0), true);
			self.scratchAmount = Utils.getNoNil(getXMLFloat(savegame.xmlFile, savegame.key.."#scratchAmount"), 0);
		else
			self:setDirtAmount(0, true);
			self.scratchAmount = 0;
		end;
	end;
end;

---Called on client side on join
-- @param integer streamId streamId
-- @param integer connection connection
-- @includeCode
function Wearable:readStream(streamId, connection)
	if self.washableNodes ~= nil then
		local dirtAmount = streamReadInt8(streamId) / 100;
		self.sentDirtAmount = dirtAmount;
		self:setDirtAmount(dirtAmount, true);
	end;
end;

---Called on server side on join
-- @param integer streamId streamId
-- @param integer connection connection
-- @includeCode
function Wearable:writeStream(streamId, connection)
	if self.washableNodes ~= nil then
		streamWriteInt8(streamId, math.floor(self:getDirtAmount()*100));
	end;
end;

---Called on on update
-- @param integer streamId stream ID
-- @param integer timestamp timestamp
-- @param table connection connection
-- @includeCode
function Wearable:readUpdateStream(streamId, timestamp, connection)
	if connection:getIsServer() and self.washableNodes ~= nil then
		local dirty = streamReadBool(streamId);
		if dirty then
			local dirtAmount = streamReadInt8(streamId) / 100;
			self:setDirtAmount(dirtAmount, true);
		end;
	end;
end;

---Called on on update
-- @param integer streamId stream ID
-- @param table connection connection
-- @param integer dirtyMask dirty mask
-- @includeCode
function Wearable:writeUpdateStream(streamId, connection, dirtyMask)
	if not connection:getIsServer() and self.washableNodes ~= nil then
		if streamWriteBool(streamId, bitAND(dirtyMask, self.washableDirtyFlag) ~= 0) then
			streamWriteInt8(streamId, math.floor(self.sentDirtAmount * 100));
		end;
	end;
end;

---Called on deleting
-- @includeCode
function Wearable:delete()
	for _, trigger in pairs(self.washingTriggers) do
		trigger:onVehicleDeleted(self);
	end;
end;

function Wearable:mouseEvent(posX, posY, isDown, isUp, button)
end;

function Wearable:keyEvent(unicode, sym, modifier, isDown)
end;

function Wearable:update(dt)
end;

---Called on update tick
-- @param float dt time since last call in ms
-- @includeCode
function Wearable:updateTick(dt)
	if self.washableNodes ~= nil then
		if self.isServer then
			if g_currentMission.environment.lastRainScale > 0.1 and g_currentMission.environment.timeSinceLastRain < 30 then
				local amount = self:getDirtAmount();
				if amount > 0.5 then
					amount = self:getDirtAmount() - (dt / self.washDuration);
				end;
				self:setDirtAmount(amount);
				self.scratchMultiplier = 2;
			else
				if self:getIsActive() or self.isActive then
					self:setDirtAmount(self:getDirtAmount() + (dt * self.dirtDuration) * self:getDirtMultiplier() * Wearable.getIntervalMultiplier());
				end;
				self.scratchMultiplier = 1;
			end;
			
			if(self.age > 1) then
				self.scratchAmount = math.min(self.scratchAmount + (dt * (self.scratchDuration * math.max(self:getDirtAmount() * 10 * 2, 1)) * self.scratchMultiplier), 1);
			else
				self.scratchAmount = 0;
			end;
			if self.washableNodes ~= nil then
				for _, node in pairs(self.washableNodes) do
					local _, y, z, w = getShaderParameter(node, "RDT");
					setShaderParameter(node, "RDT", self.scratchAmount, y, z, w, false);
				end;
			end;
		end;
	end;
end;

function Wearable:draw()
end;

---Set dirt amount
-- @param float dirtAmount new dirt amount [0..1]
-- @param boolean force force action
-- @includeCode
function Wearable:setDirtAmount(dirtAmount, force)
	if self.washableNodes ~= nil then
		self.dirtAmount = Utils.clamp(Utils.getNoNil(dirtAmount, self.dirtAmount), 0, 1);
		
		-- only update shader if dirt change sum is about 1%
		if math.abs(self.sentDirtAmount - self.dirtAmount) > 0.01 or force then
			for _, node in pairs(self.washableNodes) do
				local x,_,z,w = getShaderParameter(node, "RDT");
				setShaderParameter(node, "RDT", x, self.dirtAmount, z, w, false);
			end;
			
			if self.isServer then
				-- use min threshold (0.01) for sync limitation
				self:raiseDirtyFlags(self.washableDirtyFlag);
				self.sentDirtAmount = self.dirtAmount;
			end;
		end;
	end;
end;

---Add node to washable nodes
-- @param integer node id of node to add
-- @includeCode
function Wearable:addWashableNode(node)
	if self.washableNodes ~= nil and node ~= nil then
		Utils.getNodesByShaderParam(node, "RDT", self.washableNodes);
	end;
end;

---Remove node from washable nodes
-- @param integer node id of node to remove
-- @includeCode
function Wearable:removeWearableNode(node)
	if self.washableNodes ~= nil and node ~= nil then
		self.washableNodes[node] = nil;
	end;
end;

---Returns current dirt amount
-- @return float dirtAmount dirt amount
-- @includeCode
function Wearable:getDirtAmount(superFunc)
	return self.dirtAmount;
end;

---Returns current dirt multiplier
-- @return float dirtMultiplier current dirt multiplier
-- @includeCode
function Wearable:getDirtMultiplier(superFunc)
	local multiplier = 1;
	if self:getLastSpeed() < 1 then
		multiplier = 0;
	end;
	
	if self:getIsOnField() then
		multiplier = multiplier * self.fieldMultiplier;
	end;
	
	if superFunc ~= nil then
		multiplier = multiplier + superFunc(self);
	end;
	return multiplier;
end;

---Returns attributes and nodes to save
-- @param table nodeIdent node ident
-- @return string attributes attributes
-- @return string nodes nodes
-- @includeCode
function Wearable:getSaveAttributesAndNodes(nodeIdent)
	local attributes = 'dirtAmount="'.. tostring(self.dirtAmount) .. '" scratchAmount="' .. self.scratchAmount .. '"';
	return attributes, nil;
end;

---Add trigger to washing triggers
-- @param table trigger trigger to add
-- @includeCode
function Wearable:addWashingTrigger(trigger)
	table.insert(self.washingTriggers, trigger);
end;

---Remove trigger from washing triggers
-- @param table trigger trigger to remove
-- @includeCode
function Wearable:removeWashingTrigger(trigger)
	for i=1, table.getn(self.washingTriggers) do
		if self.washingTriggers[i] == trigger then
			table.remove(self.washingTriggers, i);
			break;
		end;
	end;
end;

---Returns dirt interval multiplier set by user
-- @return float multiplier multiplier
-- @includeCode
function Wearable.getIntervalMultiplier()
	if g_currentMission.missionInfo.dirtInterval == 1 then
		return 0;
	elseif g_currentMission.missionInfo.dirtInterval == 2 then
		return 0.25;
	elseif g_currentMission.missionInfo.dirtInterval == 3 then
		return 0.5;
	elseif g_currentMission.missionInfo.dirtInterval == 4 then
		return 1;
	end;
end;
